import sys
import regex as re
import lxml.etree as ET
from io import StringIO

from . rules import tokenize


amp_pattern = re.compile(r'&(amp;)?')
lt_pattern = re.compile(r'<|(&lt;)')
gt_pattern = re.compile(r'>|(&gt;)')


def create_tei_tree():
    root = ET.Element('TEI')
    root.set('xmlns', 'http://www.tei-c.org/ns/1.0')
    #root.set('xml:lang', 'sl')
    attr = root.attrib
    attr['{http://www.w3.org/XML/1998/namespace}lang'] = "sl"
    root.append(ET.Element('text'))
    return ET.ElementTree(root)


def index_of(string, substring, from_idx, val):
    val[0] = ''
    pattern = None
    if substring == '&amp;':
        pattern = amp_pattern
    elif substring == '&lt;':
        pattern = lt_pattern
    elif substring == '&gt;':
        pattern = gt_pattern
   
    if pattern:
        for match in pattern.finditer(string):
            if match.start() < from_idx:
                continue
            elif match.start() >= from_idx:
                val[0] = match.group()
                return match.start()
        return -1
    val[0] = substring
    return string.find(substring, from_idx)


def preprocess_tokens(tokens):
    stoken_regex = re.compile(r'<s>.*?</s>')

    org_text = []
    for match in stoken_regex.finditer(tokens):
        val = match.group()
        val = val.replace('<s>', '')
        val = val.replace('</s>', '')
        val = val.replace('<w>', '')
        val = val.replace('</w>', '')
        val = val.replace('<c>', '')
        val = val.replace('</c>', '')
        val = val.replace('<S/>', ' ')
        val = val.replace(' {2,}', ' ')
        val = re.sub(' +', ' ', val)
        org_text.append(val.strip())

    return org_text

def process_tokenize_only(para, np, os):
    if para.startswith(u"\uFEFF"):
        para = para[1:]

    tokens = tokenize(para)
    org_text = preprocess_tokens(tokens)

    token_regex = re.compile(r'<S/>|</?s>|<([wc])>([^<]+)</[wc]>')
    idx = 0
    ns = 1
    nt = 0
    old_ns = 1
    has_output = False
    for match in token_regex.finditer(tokens):
        val = match.group()
        if val == '<s>':
            nt = 0
            if ns != old_ns:
                os.write('\n')
                old_ns = ns
        elif val == '</s>':
            ns += 1
        elif val == '<S/>':
            pass
        else:
            val = match.group(2)
            actual_val = ['']
            idx_of_token = index_of(para, val, idx, actual_val)
            if idx_of_token == -1:
                print('Warning: Cannot compute token index. Token: "{}" Text: "{}"'.format(val, para))
            idx = max(idx, idx_of_token + len(actual_val[0]))
            idx_of_token += 1
            nt += 1
            line = str(np) + '.' + str(ns) + '.' + str(nt) + '.' + str(idx_of_token) + '-' + \
                    str(idx_of_token + len(actual_val[0]) - 1) + '\t' + actual_val[0] + '\n'
            os.write(line)
            has_output = True

    if has_output:
        os.write('\n')



def process_conllu(para, np, os):
    if para.startswith(u"\uFEFF"):
        para = para[1:]

    para_concat = ''.join(para)

    tokens = tokenize(para)

    os.write('# newpar id = {}\n'.format(np))

    org_text = preprocess_tokens(tokens)

    token_regex = re.compile(r'<S/>|</?s>|<([wc])>([^<]+)</[wc]>')
    idx = 0
    ns = 1
    nt = 0
    old_ns = 1
    has_output = False
    for match in token_regex.finditer(tokens):
        val = match.group()
        if val == '<s>':
            nt = 0
            if ns != old_ns:
                os.write('\n')
                old_ns = ns
            os.write('# sent_id = {}.{}\n'.format(np, ns))
            os.write('# text = {}\n'.format(org_text[ns - 1]))
        elif val == '</s>':
            ns += 1
        elif val == '<S/>':
            pass
        else:
            val = match.group(2)
            actual_val = ['']
            idx_of_token = index_of(para, val, idx, actual_val)
            if idx_of_token == -1:
                print('Warning: Cannot compute token index. Token: "{}" Text: "{}"'.format(val, para))
            idx = max(idx, idx_of_token + len(actual_val[0]))
            idx_of_token += 1
            nt += 1
            line = str(nt) + '\t{}\t_\t_\t_\t_\t_\t_\t_'.format(actual_val[0])
            if idx < len(para) and not para_concat[idx].isspace():
                space_after = 'SpaceAfter=No'
            else:
                space_after = '_'
            line += '\t{}\n'.format(space_after)
            os.write(line)
            has_output = True

    if has_output:
        os.write('\n')


def process_tei(para, np, os, tei_root):
    if para.startswith(u"\uFEFF"):
        para = para[1:]

    tokens = tokenize(para)

    parent_map = {}

    id_prefix = 'F'
    parent_node = tei_root.find('text')
    node = ET.Element('p')
    parent_node.append(node)
    #node.set('xml:id', id_prefix + str(np))
    node.attrib['{http://www.w3.org/XML/1998/namespace}id'] = id_prefix + str(np)
    parent_map[parent_node] = node

    org_text = preprocess_tokens(tokens)

    token_regex = re.compile(r'<S/>|</?s>|<([wc])>([^<]+)</[wc]>')
    idx = 0
    ns = 1
    nt = 0
    old_ns = 1
    has_output = False
    for match in token_regex.finditer(tokens):
        val = match.group()
        if val == '<s>':
            nt = 0
            node = ET.Element('s')
            #node.set('xml:id', id_prefix + str(np) + '.' + str(ns))
            node.attrib['{http://www.w3.org/XML/1998/namespace}id'] = id_prefix + str(np) + '.' + str(ns)
            parent_node.append(node)
            parent_map[node] = parent_node
            parent_node = parent_node[-1]
        elif val == '</s>':
            parent_node = parent_map[parent_node]
            ns += 1
        elif val == '<S/>':
            node = ET.Element('c')
            node.text = ' '
            parent_node.append(node)
            parent_map[node] = parent_node
        else:
            val = match.group(2)
            actual_val = ['']
            idx_of_token = index_of(para, val, idx, actual_val)
            if (idx_of_token == -1):
                print('Warning: cannot compute token index. Token: "{}" Text "{}"'.format(val, para))
            idx = max(idx, idx_of_token + len(actual_val[0]))
            idx_of_token += 1
            nt += 1
            if match.group(1) == 'c':
                tag_name = 'pc'
            else:
                tag_name = 'w'
            node = ET.Element(tag_name)
            node.text = actual_val[0]
            #node.set('xml:id', id_prefix + str(np) + '.' + str(ns) + '.t' + str(nt))
            node.attrib['{http://www.w3.org/XML/1998/namespace}id'] = id_prefix + str(np) + '.' + str(ns) + '.t' + str(nt)
            parent_node.append(node)
            parent_map[node] = parent_node


def process_text(text, os, tei_root, conllu, pass_newdoc_id):
    np = 0
    for line in text:
        if line.isspace() or line == '':
            continue

        # Normalize exotic characters
        line = normalize(line)

        if conllu:
            if pass_newdoc_id and line.startswith('# newdoc id = '):
                np = 0
                os.write(line)
                os.write('\n')
            np += 1
            process_conllu(line, np, os)
        elif tei_root is not None:
            np += 1
            process_tei(line, np, os, tei_root)
        else:
            np += 1
            process_tokenize_only(line, np, os)

def normalize(text):
    text = text.replace('\xad', '-')   # Soft hyphens
    return text


def run(text=None, in_file=None, in_files=None, out_file=None, to_stdout=False, tei=False, conllu=False, pass_newdoc_id=False):
    """
    Run Obeliks on specified input.

    Reads input from 'text' parameter, or from files passed via 'in_file' or 'in_files'. If none of these three 
    parameters are supplied, input is read from stdin. The output gets written to the path specified inside
    'out_file' or to stdout if 'to_stdout' is True. If 'out_file' is None and 'to_stdout is False, return output
    as a string.

    Args:
        text: A string containing input text.
        in_file: Path to file with input text.
        in_files: List of files with input text.
        out_file: Path where the output gets written to.
        to_stdout: Write output to stdout. Has priority over 'out_file'.
        tei: Output in XML-TEI format.
        conllu: Output in CoNLL-U format.
        pass_newdoc_id: Pass lines starting with \"# newdoc id =\" to output. Only applies if CoNLL-U output is 
                        selected.
    """
    os = StringIO()
    if to_stdout:
        os = sys.stdout
    elif out_file is not None:
        os = open(out_file, 'w', encoding = 'utf-8')

    if in_files and len(in_files) > 0:
        text = []
        for file_name in in_files:
            with open(file_name) as f:
                text += f.readlines()
    elif in_file is not None:
        with open(in_file) as f:
            text = f.readlines()
    elif text:
        text = text.splitlines()
    else:
        text = sys.stdin.readlines()

    if tei:
        tei_tree = create_tei_tree()
        process_text(text, os, tei_tree.getroot(), False, False)
        if isinstance(os, StringIO):
            return ET.tostring(tei_tree, encoding='utf8', method='xml')
        else:
            tei_tree.write(os.buffer, encoding='utf-8', xml_declaration=True, pretty_print=True)
    else:
        process_text(text, os, None, conllu, pass_newdoc_id)
    
    if isinstance(os, StringIO):
        contents = os.getvalue()
        os.close()
        return contents

    if os is not sys.stdout:
        os.close()

    return None




import regex as re
from pathlib import Path


tag_regex = re.compile(r'\</?[^>]+\>')
abbrv_regex = re.compile(r'<w>(\p{L}+)<\/w><c>\\.<\/c>')
endofsentence_regex = re.compile(r'^<[wc]>[\p{Lu}"»“‘\'0-9]$')
abbrvexcl_regex = re.compile(r'(?P<step><w>(?P<word>\p{L}+)<\/w><c>\.<\/c>(?P<tail><S\/>)?)(?P<ctx>(<\/[ps]>)|(<[wc]>.))')
abbrvother_regex = re.compile(r'(?P<step><w>(?P<word>\p{L}+)<\/w><c>\.<\/c>(?P<tail><S\/>)?)<[wc]>[:,;0-9\p{Ll}]')

def load_list(path):
    tokens = set()
    with open(path, 'r') as f:
        for token in f:
            token = token.strip()
            if not token == '' and not token.startswith('#'):
                tokens.add(token)
    return tokens

res_path = Path(__file__).parent / 'res/'

abbrv_seq = load_list(res_path / 'ListOSeq.txt')
abbrv_seg_seq = load_list(res_path / 'ListOSegSeq.txt')
abbrv_no_seg_seq = load_list(res_path / 'ListONoSegSeq.txt')
abbrv_excl = load_list(res_path / 'ListOExcl.txt')
abbrv_excl_CS = load_list(res_path / 'ListOExclCS.txt')
abbrv_seg = load_list(res_path / 'ListOSeg.txt')
abbrv_all = load_list(res_path / 'ListOAll.txt')
abbrv_all_CS = load_list(res_path / 'ListOAllCS.txt')

abbrv_seq_len = []
for abbrv in abbrv_seq:
    length = 0
    for c in abbrv:
        if c == '.':
            length += 1
        abbrv_seq_len.append(length)
abbrv_seq_len.sort(reverse = True)


def empty_if_none(s):
    if s is None:
        return ''
    else:
        return s.strip()


def process_abbrv_seq(text, seq_len):
    idx = 0
    regex = re.compile(r'(?P<jump>(?P<step><w>\p{L}+</w><c>\\.</c>(<S/>)?)(<w>\p{L}+</w><c>\\.</c>(<S/>)?){' \
            + str(seq_len - 1) + r'})(?P<ctx>(</[ps]>)|(<[wc]>.))')

    sb = []
    while True:
        m = regex.search(text, idx)
        if not m:
            break
        sb.append(text[idx : m.start()])
        xml = empty_if_none(m.group('jump'))
        abbrv_lower = re.sub(tag_regex, '', xml).sub(' ', '').lower()
        if abbrv_lower in abbrv_seq:
            idx = m.start() + len(xml)
            xml = re.sub(abbrv_regex, '<w>$1.</w>', xml)
            if endofsentence_regex.search(empty_if_none(m.group('ctx'))):
                if abbrv_lower in abbrv_seg_seq:
                    xml += '</s><s>'
                elif abbrv_lower in abbrv_no_seg_seq:
                    xml += '<!s/>'
        else:
            xml = empty_if_none(m.group('step'))
            idx = m.start() + len(xml)
        sb.append(xml)

    sb.append(text[idx:])
    return ''.join(sb)


def process_abbrv_excl(text):
    idx = 0
    sb = []
    while True:
        m = abbrvexcl_regex.search(text, idx)
        if not m:
            break
        sb.append(text[idx : m.start()])
        word = empty_if_none(m.group('word'))
        word_lower = word.lower()
        if len(word) == 1 or word_lower in abbrv_excl or word in abbrv_excl_CS:
            xml = '<w>' + word + '.</w>' + empty_if_none(m.group('tail'))
            idx = m.start() + len(empty_if_none(m.group('step')))
            if word_lower in abbrv_seg and endofsentence_regex.search(empty_if_none(m.group('ctx'))):
                xml += '</s><s>'
        else:
            xml = empty_if_none(m.group('step'))
            idx = m.start() + len(xml)
        sb.append(xml)

    sb.append(text[idx:])
    return ''.join(sb)


def process_abbrv_other(text):
    idx = 0
    sb = []
    while True:
        m = abbrvother_regex.search(text, idx)
        if not m:
            break
        sb.append(text[idx : m.start()])
        word = empty_if_none(m.group('word'))
        word_lower = word.lower()
        if word_lower in abbrv_all or word in abbrv_all_CS:
            xml = '<w>' + word + '.</w>' + empty_if_none(m.group('tail'))
            idx = m.start() + len(empty_if_none(m.group('step')))
        else:
            xml = empty_if_none(m.group('step'))
            idx = m.start() + len(xml)
        sb.append(xml)
   
    sb.append(text[idx:])
    return ''.join(sb)


def exec_rules(text, rules):
    for rule in rules:
        regex = rule[0]    # Compiled search regex pattern
        rhs = rule[1]      # Replacement string

        has_val = '$val' in rhs
        has_txt = '$txt' in rhs

        if not has_val and not has_txt:
            text = re.sub(regex, rhs, text)
        else:
            def repl_func(match):
                if has_val:
                    return rhs.replace('$val', match.group(0))
                if has_txt:
                    return rhs.replace('$txt', re.sub(tag_regex, '', match.group(0)))
            text = re.sub(regex, repl_func, text) 

    return text


def load_rules(path):
    rules = []
    split_regex = re.compile(r'^(?P<regex>.*)((--)|(==))\>(?P<rhs>.*)$')

    with open(path, 'r') as f:
        for line in f:
            line = line.strip()
            if not line.startswith('#') and len(line) > 0:
                #opt = 0
                opt = re.UNICODE
                if '-->' in line:
                   opt |= re.IGNORECASE 

                match = re.search(split_regex, line)
                if match:
                    regex = re.compile(empty_if_none(match.group('regex')), opt)
                    rhs = empty_if_none(match.group('rhs'))

                    rules.append((regex, rhs))
    return rules


def tokenize(text):
    tok_rules_1 = load_rules(res_path / 'TokRulesPart1.txt')
    tok_rules_2 = load_rules(res_path / 'TokRulesPart2.txt')

    xml =  exec_rules(text, tok_rules_1)
    for seq_len in abbrv_seq_len:
        xml = process_abbrv_seq(xml, seq_len)
    xml = process_abbrv_excl(xml)
    xml = process_abbrv_other(xml)
    xml = exec_rules(xml, tok_rules_2)
    xml = xml.replace('<!s/>', '')

    return xml

